--
-- mogliBasics
-- This is the specialization for mogliBasics
--

if mogliBase11 == nil then
	mogliBase11 = {}

	mogliBase11.version = 1.1
	mogliBase11.sendLocal = false

	function mogliBase11.newSubclass()
		local subclass = {}
		setmetatable( subclass, { __index = function (table, key) return mogliBase11[key] end } )
		subclass.baseDirectory = g_currentModDirectory
		subclass.modsDirectory = g_modsDirectory.."/"
		return subclass
	end

	function mogliBase11.globalsReset( createIfMissing )
	end

	function mogliBase11.globalsLoad( file , rootTag, globals )	

		local xmlFile = loadXMLFile( "mogliBasics", file, rootTag )
		for name,value in pairs(globals) do
			local tp = getXMLString(xmlFile, rootTag.."." .. name .. "#type")
			if     tp == nil then
	--			print(file..": "..name.." = nil")
			elseif tp == "bool" then
				local bool = getXMLBool( xmlFile, rootTag.."." .. name .. "#value" )
				if bool ~= nil then
					--if bool then globals[name] = 1 else globals[name] = 0 end
					globals[name] = bool
				end
	--			print(file..": "..name.." = "..tostring(globals[name]))
			elseif tp == "float" then
				local float = getXMLFloat( xmlFile, rootTag.."." .. name .. "#value" )
				if float ~= nil then globals[name] = float end
	--			print(file..": "..name.." = "..tostring(globals[name]))
			elseif tp == "int" then
				local int = getXMLInt( xmlFile, rootTag.."." .. name .. "#value" )
				if int ~= nil then globals[name] = int end
	--			print(file..": "..name.." = "..tostring(globals[name]))
			else
				print(file..": "..name..": invalid XML type : "..tp)
			end
		end
	end

	function mogliBase11.getText(id)
		if id == nil then
			return "nil";
		end;
		
		if g_i18n:hasText( id ) then
			return g_i18n:getText( id )
		end
		
		return id
	end;

	function mogliBase11.normalizeAngle( angle )
		local normalizedAngle = angle
		if angle > math.pi then
			normalizedAngle = angle - math.pi - math.pi
		elseif angle <= -math.pi then
			normalizedAngle = angle + math.pi + math.pi
		end
		return normalizedAngle
	end

	function mogliBase11.prerequisitesPresent(specializations)
		return true
	end

	function mogliBase11:load(xmlFile)
	-- should always be overwritten
		mogliBase11.registerState( self, "mogliBasicsDummy", false, mogliBase11.debugEvent )
	end

	function mogliBase11:delete()
	end

	function mogliBase11:mouseEvent(posX, posY, isDown, isUp, button)
	end

	function mogliBase11:keyEvent(unicode, sym, modifier, isDown)
	end

	function mogliBase11:update(dt)
	end

	function mogliBase11:updateTick(dt)	
	end

	function mogliBase11:readStream(streamId, connection)
	end

	function mogliBase11:writeStream(streamId, connection)
	end

	function mogliBase11:draw()	
	end  
				 
	function mogliBase11:getSaveAttributesAndNodes(nodeIdent)
		local attributes = ""
		return attributes
	end;

	function mogliBase11:loadFromAttributesAndNodes(xmlFile, key, resetVehicles)
		return BaseMission.VEHICLE_LOAD_OK;
	end

	function mogliBase11:initStateHandling( )
		if self.mbVersion11 == nil or self.mbVersion11 < mogliBase11.version then
			self.mbVersion11 = mogliBase11.version
			if self.mbSetState11 ~= mogliBase11.mbSetState then
				self.mbSetState11 = mogliBase11.mbSetState
			end
			if self.mbStateHandler11 == nil then
				self.mbStateHandler11 = {}
			end
		end
	end

	function mogliBase11.getValueType( value )

		if value == nil then
			return "string"
		end

		local valueType = type( value )
		
		if valueType == "number" then
			local n = value
			if n - math.floor( n ) < 1E-3 then
				valueType = "int8"
			else
				valueType = "float32"
			end
		end
		
		return valueType
	end

	function mogliBase11:registerState( level1, default, handler, valueType )
		mogliBase11.initStateHandling( self )
		if default ~= nil then
			self[level1] = default
		end
		local valueTypeShort
		self.mbStateHandler11[level1] = { handler = handler, valueType = Utils.getNoNil( valueType, mogliBase11.getValueType( default ) ), default = default }
		table.sort( self.mbStateHandler11 )
	end

	function mogliBase11:mbSetState(level1, value, noEventSend)
		mogliBase11.initStateHandling( self )
		
		if self.mbStateHandler11[level1] == nil then
			mogliBase11.registerState( self, level1, nil, nil, mogliBase11.getValueType( value ) )
		end
		
		if self[level1] == nil or self[level1] ~= value then			
			local valueType = self.mbStateHandler11[level1].valueType
			if noEventSend == nil or noEventSend == false then
				if g_server ~= nil then 
					g_server:broadcastEvent(mogliBase11Event:new(self, level1, valueType, value), mogliBase11.sendLocal, nil, object) 
				else
					g_client:getServerConnection():sendEvent(mogliBase11Event:new(self, level1, valueType, value)) 
				end 
			end 						
		
			if self.mbStateHandler11[level1].handler ~= nil then
				local noEventSend2 = true 
				if self.isServer then
					noEventSend2 = noEventSend 
				end 
				
				local state, message = pcall( self.mbStateHandler11[level1].handler, self, self[level1], value, noEventSend2 )
				if not state then
					print("Error: "..tostring(message)) 
					mogliBase11.debugEvent( self, self[level1], value, noEventSend ) 
					self[level1] = value 	
				end
			else
				self[level1]= value 	
			end 	
		end 	
	end 

	local function writeStreamTypedValue( streamId, valueType, value )
		if     valueType == "string"  then
			streamWriteString(streamId,self[level1]) 
		elseif valueType == "int8"    then
			streamWriteInt8(streamId,self[level1]) 
		elseif valueType == "float32" 
				or valueType == "number"  then
			streamWriteFloat32(streamId,self[level1])  
		elseif valueType == "boolean" then
			streamWriteBool(streamId,self[level1]) 
		else 
			streamWriteString(streamId,tostring(self[level1]))
		end		
	end

	local function readStreamTypedValue( streamId, valueType )
		local value = nil
		if     valueType == "string"  then
			value = streamReadString(streamId) 
		elseif valueType == "int8"    then
			value = streamReadInt8(streamId) 
		elseif valueType == "float32" 
				or valueType == "number"  then
			value = streamReadFloat32(streamId) 
		elseif valueType == "boolean" then
			value = streamReadBool(streamId) 
		else
			value = streamReadString(streamId) 
		end
		return value
	end 
	
	function mogliBase11:readStream(streamId, connection)
		for level1,state in pairs( self.mbStateHandler11 ) do
			local value = readStreamTypedValue( streamId, state.valueType )
			if value ~= nil then			
				mogliBase11:mbSetState(level1, value, true )
			end
		end
	end 
	 
	function mogliBase11:writeStream(streamId, connection)
		for level1,state in pairs( self.mbStateHandler11 ) do
			writeStreamTypedValue( streamId, state.valueType, self[level1] )
		end
	end 


	function mogliBase11:debugEvent( old, new, noEventSend )
		local i = 2 
		local info 
		print("------------------------------------------------------------------------") 
		while i <= 10 do
			info = debug.getinfo(i) 
			if info == nil then break end
			print(string.format("%i: %s (%i): %s", i, info.short_src, Utils.getNoNil(info.currentline,0), Utils.getNoNil(info.name,"<???>"))) 
			i = i + 1 
		end
		if info ~= nil and info.name ~= nil and info.currentline ~= nil then
			print("...") 
		end
		print("------------------------------------------------------------------------") 
		print(tostring(old).." "..tostring(new).." "..tostring(noEventSend)) 
	end 


	mogliBase11Event = {} 
	mogliBase11Event_mt = Class(mogliBase11Event, Event) 

	InitEventClass(mogliBase11Event, "mogliBase11Event") 

	 
	function mogliBase11Event:emptyNew()
		local self = Event:new(mogliBase11Event_mt) 
		self.className="mogliBase11Event" 
		return self 
	end 

	function mogliBase11Event:new(object, level1, valueType, value)
		local self = mogliBase11Event:emptyNew() 
		self.object    = object 
		self.level1    = level1 
		self.valueType = valueType
		self.value     = value 
		return self 
	end 

	function mogliBase11Event:readStream(streamId, connection)
		--both clients and server can receive this event
		local id = streamReadInt32(streamId) 
		self.object = networkGetObject(id) 
		self.level1    = streamReadString(streamId) 
		self.valueType = streamReadString(streamId) 
		
		if     type(self.valueType) ~= "string" 
				or self.valueType       == ""
				or string.sub(self.valueType,1,1) == "s" then
			self.valueType = "string"
		elseif string.sub(self.valueType,1,1) == "n" then
			self.valueType = "number"
		elseif string.sub(self.valueType,1,1) == "f" then
			self.valueType = "float32"
		elseif string.sub(self.valueType,1,1) == "i" then
			self.valueType = "int8"
		elseif string.sub(self.valueType,1,1) == "b" then
			self.valueType = "boolean"
		end 
		
		if mogliBase11.sendLocal then 
			print("readStream: "..tostring(self.value).." ("..self.valueType..")")
		end
		
		self.value     = readStreamTypedValue( streamId, self.valueType )
		self:run(connection) 
	end 

	function mogliBase11Event:writeStream(streamId, connection)
		--both clients and server can send this event
		streamWriteInt32(streamId, networkGetObjectId(self.object)) 
		streamWriteString(streamId, self.level1 ) 
		
		if     type(self.valueType) ~= "string" 
				or self.valueType       == "" then
			streamWriteString(streamId, "s")
		else
			streamWriteString(streamId, string.sub(self.valueType,1,1))
		end 				
		
		writeStreamTypedValue( streamId, self.valueType, self.value )

		if mogliBase11.sendLocal then 
			print("writeStream: "..tostring(self.value).." ("..self.valueType..")")
		end		
	end 

	function mogliBase11Event:run(connection)
		----both clients and server can "run" this event (after reading it)	
		mogliBase11.mbSetState(self.object, self.level1, self.value, true) 
		if not connection:getIsServer() then  
			g_server:broadcastEvent(mogliBase11Event:new(self.object, self.level1, self.valueType, self.value), nil, connection, self.object) 
		end 
	end 
end