--***************************************************************
--
-- based on dural's mrGearbox2.lua
-- 
-- version 1.3 by mogli (biedens)
-- 2014/08/06
--
--***************************************************************

source(Utils.getFilename("mogliBase.lua", g_currentModDirectory))
mrGearboxMogli = mogliBase11.newSubclass()
mrGearboxMogli.huge = 1000000

function mrGearboxMogli.prerequisitesPresent(specializations) 
	return true
end 

function mrGearboxMogli:load(xmlFile) 
	local key = "vehicle.gearboxMogli"
	if hasXMLProperty(xmlFile, key) then
		mrGearboxMogli.initFromXml(self,xmlFile,key)
	end
end

function mrGearboxMogli.getNoNil2( v0, v1, v2, use2 )
	if v0 == nil then
		if v2 ~= nil and use2 then
			return v2
		end
		return v1
	end
	return v0
end

function mrGearboxMogli:initFromXml(xmlFile,xmlString) 

	self.mrGbMSetGear           = mrGearboxMogli.mrGbMSetGear 
	self.mrGbMSetRange          = mrGearboxMogli.mrGbMSetRange 
	self.mrGbMSetRange2         = mrGearboxMogli.mrGbMSetRange2
	self.mrGbMSetAutomatic      = mrGearboxMogli.mrGbMSetAutomatic 
	self.mrGbMGetClutchPercent  = mrGearboxMogli.mrGbMGetClutchPercent 
	self.mrGbMDoGearShift       = mrGearboxMogli.mrGbMDoGearShift 

	self.mrGbMConfigVersion           = Utils.getNoNil(getXMLFloat(xmlFile, xmlString .. "#version" ),1.3)
	self.mrGbMDefaultOn               = Utils.getNoNil(getXMLBool( xmlFile, xmlString .. "#defaultOn" ),true)
	self.mrGbMSwapGearRangeKeys       = Utils.getNoNil(getXMLBool( xmlFile, xmlString .. "#swapGearRangeKeys" ),false)
	
	self.mrGbMRatedRpm                = Utils.getNoNil(getXMLFloat(xmlFile, xmlString .. "#ratedRpm"), self.motor.maxRpm  )  --0.86* self.motor.minRpm) / self.mrGbMRatedRpm  
	self.mrGbMIdleRpm			            = Utils.getNoNil(getXMLFloat(xmlFile, xmlString .. "#idleRpm"),  self.motor.minRpm  )  
	self.mrGbMIdleRpmRatio            = self.motor.minRpm / self.mrGbMRatedRpm  
	self.mrGbMOpenRpm                 = Utils.getNoNil(getXMLFloat(xmlFile, xmlString .. "#clutchOpenRpm"), self.mrGbMIdleRpm-50) -- no automatic opening of clutch by default!!!
	self.mrGbMCloseRpm                = Utils.getNoNil(getXMLFloat(xmlFile, xmlString .. "#clutchCloseRpm"), self.mrGbMIdleRpm+100 )
	self.mrGbMAutoStartStop           = Utils.getNoNil(getXMLBool( xmlFile, xmlString .. "#autoStartStop"), true)
	
	local clutchEngagingTimeMs = getXMLFloat(xmlFile, xmlString .. "#clutchEngagingTimeMs")
	if clutchEngagingTimeMs ~= nil then
		self.mrGbMClutchTimeInc        = Utils.getNoNil(getXMLFloat(xmlFile, xmlString .. "#clutchTimeIncreaseMs"), clutchEngagingTimeMs )  -- 1000 = 1s
		self.mrGbMClutchTimeDec        = Utils.getNoNil(getXMLFloat(xmlFile, xmlString .. "#clutchTimeDecreaseMs"), clutchEngagingTimeMs )  -- 1000 = 1s
	else
		self.mrGbMClutchTimeInc        = Utils.getNoNil(getXMLFloat(xmlFile, xmlString .. "#clutchTimeIncreaseMs"), 2000)  -- 1000 = 1s
		if self.mrGbMOpenRpm > 0 then clutchEngagingTimeMs = 1000 else clutchEngagingTimeMs = self.mrGbMClutchTimeInc end
		self.mrGbMClutchTimeDec        = Utils.getNoNil(getXMLFloat(xmlFile, xmlString .. "#clutchTimeDecreaseMs"), clutchEngagingTimeMs)  -- 1000 = 1s
	end
	
	self.mrGbMClutchShiftTime         = Utils.getNoNil(getXMLFloat(xmlFile, xmlString .. "#clutchShiftingTimeMs"), 0.25 * self.mrGbMClutchTimeDec) 
	self.mrGbMTransmissionEfficiency  = Utils.getNoNil(getXMLFloat(xmlFile, xmlString .. "#transmissionEfficiency"), 0.93) / 0.88
	self.mrGbMTransmissionEfficiencyDec=Utils.getNoNil(getXMLFloat(xmlFile, xmlString .. "#transmissionEfficiencyDec"), 0.07) / 0.88
		
	local alwaysDoubleClutch          = Utils.getNoNil(getXMLBool(xmlFile, xmlString .. "#doubleClutch"), false) 
	self.mrGbMGearsDoubleClutch       = Utils.getNoNil(getXMLBool(xmlFile, xmlString .. ".gears#doubleClutch"), alwaysDoubleClutch) 
	self.mrGbMRange1DoubleClutch      = Utils.getNoNil(getXMLBool(xmlFile, xmlString .. ".ranges(0)#doubleClutch"), alwaysDoubleClutch) 
	self.mrGbMRange2DoubleClutch      = Utils.getNoNil(getXMLBool(xmlFile, xmlString .. ".ranges(1)#doubleClutch"), alwaysDoubleClutch) 
	self.mrGbMReverseDoubleClutch     = Utils.getNoNil(getXMLBool(xmlFile, xmlString .. ".reverse#doubleClutch"), alwaysDoubleClutch) 
	
	self.mrGbMGearTimeToShiftGear     = Utils.getNoNil(getXMLFloat(xmlFile, xmlString .. ".gears#shiftTimeMs"),      500 )
	self.mrGbMGearTimeToShiftHl       = Utils.getNoNil(getXMLFloat(xmlFile, xmlString .. ".ranges(0)#shiftTimeMs"),  800 ) 
	self.mrGbMGearTimeToShiftRanges2  = Utils.getNoNil(getXMLFloat(xmlFile, xmlString .. ".ranges(1)#shiftTimeMs"),  800 ) 
	self.mrGbMGearTimeToShiftReverse  = Utils.getNoNil(getXMLFloat(xmlFile, xmlString .. ".reverse#shiftTimeMs"),   1000 ) 

	self.mrGbMClutchAfterShiftGear    = Utils.getNoNil(getXMLFloat(xmlFile, xmlString .. ".gears#clutchRatio"), 0.7) 
	self.mrGbMClutchAfterShiftHl      = Utils.getNoNil(getXMLFloat(xmlFile, xmlString .. ".ranges(0)#clutchRatio"), 0.7) 
	self.mrGbMClutchAfterShiftRanges2 = Utils.getNoNil(getXMLFloat(xmlFile, xmlString .. ".ranges(1)#clutchRatio"), 0.7) 
	self.mrGbMClutchAfterShiftReverse = Utils.getNoNil(getXMLFloat(xmlFile, xmlString .. ".reverse#clutchRatio"), 0 ) 
	self.mrGbMMinClutchPercent        = Utils.getNoNil(getXMLFloat(xmlFile, xmlString .. "#minClutchRatio"), 0.2 )
	self.mrGbMMaxClutchPercent        = Utils.getNoNil(getXMLFloat(xmlFile, xmlString .. "#maxClutchRatio"), 1 )
	self.mrGbMClutchTorqueClosed      = Utils.getNoNil(getXMLFloat(xmlFile, xmlString .. "#clutchTorqueClosed"), self.motor.maxClutchTorque ) -- ) 10 * self.motor.maxMotorTorque )
	self.mrGbMClutchTorqueConst       = Utils.getNoNil(getXMLFloat(xmlFile, xmlString .. "#clutchTorqueConst"), self.motor.maxMotorTorque )
	self.mrGbMClutchTorqueDiff        = Utils.getNoNil(getXMLFloat(xmlFile, xmlString .. "#clutchTorqueDiff"), self.motor.maxClutchTorque - self.motor.maxMotorTorque )
	self.mrGbMClutchExponent          = Utils.getNoNil(getXMLFloat(xmlFile, xmlString .. "#clutchExponent"), 1 )
	self.mrGbMMaxClutchPercentInv     = 1 / self.mrGbMMaxClutchPercent

	self.mrGbMRealMotorBrakeFx        = Utils.getNoNil(getXMLFloat(xmlFile, xmlString .. "#motorBrakeFx"), 0.1 )
	self.mrGbMGlobalRatioFactor       = Utils.getNoNil(getXMLFloat(xmlFile, xmlString .. "#globalRatioFactor"), 1.05 )

	self.mrGbMRpmIncFactor            = ( self.mrGbMRatedRpm - self.mrGbMIdleRpm ) / Utils.getNoNil(getXMLFloat(xmlFile, xmlString .. "#revUpDownMs"), 1000 )
	self.mrGbMAutoShiftTimeout        = Utils.getNoNil(getXMLFloat(xmlFile, xmlString .. ".gears#autoShiftTimeout"), 4000 ) 
	self.mrGbMAutoShiftTimeout2       = Utils.getNoNil(getXMLFloat(xmlFile, xmlString .. ".gears#autoShiftTimeout2"),1000 ) 

	self.mrGbMOverspeedRpm            = self.motor.maxRpm + 100
		
	local reverseMinGear  = getXMLInt(xmlFile, xmlString .. ".reverse#minGear")
	local reverseMaxGear  = getXMLInt(xmlFile, xmlString .. ".reverse#maxGear")
	local reverseMinRange = getXMLInt(xmlFile, xmlString .. ".reverse#minRange")
	local reverseMaxRange = getXMLInt(xmlFile, xmlString .. ".reverse#maxRange")
	local rangeGearOffset = Utils.getNoNil(getXMLInt(xmlFile, xmlString .. ".ranges#gearOffset"), 0) 
	local gearRangeOffset = Utils.getNoNil(getXMLInt(xmlFile, xmlString .. ".gears#rangeOffset"), 0) 
		
	self.mrGbMGears = {} 
	local i = 0 
	while true do
		local baseName = xmlString .. string.format(".gears.gear(%d)", i) 		
		local speed    = getXMLFloat(xmlFile, baseName .. "#speed") 
		if speed==nil then
			break 
		end 
		i = i + 1 
		local name  = Utils.getNoNil(getXMLString(xmlFile, baseName .. "#name"), tostring(i)) 

		local fwdOnly = getXMLBool(xmlFile, baseName .. "#forwardOnly" )
		if fwdOnly == nil then
			fwdOnly = not ( ( reverseMinGear == nil or i >= reverseMinGear )
									and ( reverseMaxGear == nil or i <= reverseMaxGear ) )
		end		
		local revOnly = Utils.getNoNil( getXMLBool( xmlFile, baseName .. "#reverseOnly" ), false )
		
		local upRangeOffset   = Utils.getNoNil( getXMLInt( xmlFile, baseName .. "#upRangeOffset" ),  -gearRangeOffset ) 
		local downRangeOffset = Utils.getNoNil( getXMLInt( xmlFile, baseName .. "#downRangeOffset" ), gearRangeOffset ) 
		
		local newGear = {speed=speed/3.6,name=name,forwardOnly=fwdOnly,reverseOnly=revOnly,upRangeOffset=upRangeOffset,downRangeOffset=downRangeOffset} 
		
	--print(tostring(newGear.speed).." "..tostring(newGear.name).." "..tostring(newGear.forwardOnly).." "..tostring(newGear.reverseOnly).." "..tostring(newGear.upRangeOffset).." "..tostring(newGear.downRangeOffset))
					
		table.insert(self.mrGbMGears, newGear)  -- m/s
	end 	
	
	if i==0 then
		local newGear = {speed=self.motor.maxForwardSpeed/self.mrGbMGlobalRatioFactor,name="1",forwardOnly=false,reverseOnly=false,upRangeOffset=0,downRangeOffset=0} 
		table.insert(self.mrGbMGears, newGear)  -- m/s
	end 	
			
	self.mrGbMRanges = {} 
	i = 0 
	local generateNames = true 
	while true do
		local baseName = xmlString .. string.format(".ranges(0).range(%d)", i) 		
		local ratio = getXMLFloat(xmlFile, baseName .. "#ratio") 
		if ratio==nil then
			break 
		end 
		i = i + 1 
		
		local name = getXMLString(xmlFile, baseName .. "#name") 
		if name == nil then
			name = tostring(i) 
		else
			generateNames = false 
		end
		
		local fwdOnly = getXMLBool(xmlFile, baseName .. "#forwardOnly" )
		if fwdOnly == nil then
			fwdOnly = not ( ( reverseMinRange == nil or i >= reverseMinRange )
					        and ( reverseMaxRange == nil or i <= reverseMaxRange ) )
		end		
		local revOnly = Utils.getNoNil( getXMLBool(xmlFile, baseName .. "#reverseOnly" ), false )
		
		local upGearOffset   = Utils.getNoNil( getXMLInt( xmlFile, baseName .. "#upGearOffset" ),  -rangeGearOffset ) 
		local downGearOffset = Utils.getNoNil( getXMLInt( xmlFile, baseName .. "#downGearOffset" ), rangeGearOffset ) 
		
		local newRange = {ratio=ratio,name=name,forwardOnly=fwdOnly,reverseOnly=revOnly,upGearOffset=upGearOffset,downGearOffset=downGearOffset} 
		
	--print(tostring(newRange.ratio).." "..tostring(newRange.name).." "..tostring(newRange.forwardOnly).." "..tostring(newRange.reverseOnly).." "..tostring(newRange.upGearOffset).." "..tostring(newRange.downGearOffset))
		
		table.insert(self.mrGbMRanges, newRange)  -- m/s
	end 
	
	if i==0 then
		local newRange = {ratio=1,name="",forwardOnly=false,reverseOnly=false,upGearOffset=0,downGearOffset=0} 
		table.insert(self.mrGbMRanges, newRange)  -- m/s
		generateNames = false
	end
	
	if generateNames then
		local fwd,rev=0,0
		for _,newRange in pairs( self.mrGbMRanges ) do
			if     newRange.forwardOnly then
				fwd = fwd + 1
			elseif newRange.reverseOnly then
				rev = rev + 1
			else
				fwd = fwd + 1
				rev = rev + 1
			end
		end
		
		local fwd2,rev2=0,0

		for _,newRange in pairs( self.mrGbMRanges ) do
			local isFwd,isRev=true,true
			if     newRange.forwardOnly then
				fwd2 = fwd2 + 1
				isRev = false
			elseif newRange.reverseOnly then
				rev2 = rev2 + 1
				isFwd = false
			else
				fwd2 = fwd2 + 1
				rev2 = rev2 + 1
			end
			if     isFwd then
				if     fwd == 1 then
					newRange.name = ""
				elseif fwd == 2 then
					if fwd2 == 1 then
						newRange.name = "L"
					else
						newRange.name = "H"
					end
				elseif fwd == 3 then
					if     fwd2 == 1 then
						newRange.name = "L"
					elseif fwd2 == 2 then
						newRange.name = "M"
					else
						newRange.name = "H"
					end
				elseif fwd == 4 then
					if     fwd2 == 1 then
						newRange.name = "S"
					elseif fwd2 == 2 then
						newRange.name = "L"
					elseif fwd2 == 3 then
						newRange.name = "M"
					else
						newRange.name = "H"
					end
				else
					newRange.name = "G"..tostring(fwd2)
				end
			elseif isRev then
				if rev==1 then
					newRange.name = "R"
				else
					newRange.name = "R"..tostring(rev2)
				end			
			end
		end		
	end
	
	self.mrGbMRanges2 = {} 
	i = 0 
	while true do
		local baseName = xmlString .. string.format(".ranges(1).range(%d)", i) 		
		local ratio = getXMLFloat(xmlFile, baseName .. "#ratio") 
		if ratio==nil then
			break 
		end 
		i = i + 1 
		
		local name = getXMLString(xmlFile, baseName .. "#name") 
		if name == nil then
			name = "G"..tostring(i) 
		end
		
		local fwdOnly = Utils.getNoNil( getXMLBool(xmlFile, baseName .. "#forwardOnly" ), false )
		local revOnly = Utils.getNoNil( getXMLBool(xmlFile, baseName .. "#reverseOnly" ), false )
		
		local newRange = {ratio=ratio,name=name,forwardOnly=fwdOnly,reverseOnly=revOnly} 
		table.insert(self.mrGbMRanges2, newRange)  -- m/s
	end 
	
	if i==0 then
		local newRange = {ratio=1,name="",forwardOnly=false,reverseOnly=false} 
		table.insert(self.mrGbMRanges2, newRange)  -- m/s
	end
		
	self.mrGbMReverseRatio            = Utils.getNoNil(getXMLFloat(xmlFile, xmlString .. ".reverse#ratio"), 1) 
	
	
	
	self.mrGbMReverseResetGear        = false 
	if getXMLBool(xmlFile, xmlString .. ".reverse#resetGear") or getXMLBool(xmlFile, xmlString .. ".gears#reverseReset") then
		self.mrGbMReverseResetGear      = true 
	end
	self.mrGbMReverseResetRange       = false 
	if getXMLBool(xmlFile, xmlString .. ".reverse#resetRange") or getXMLBool(xmlFile, xmlString .. ".ranges(0)#reverseReset") then
		self.mrGbMReverseResetRange     = true
	end
	self.mrGbMReverseResetRange2      = false
	if getXMLBool(xmlFile, xmlString .. ".ranges(1)#reverseReset") then
		self.mrGbMReverseResetRange2    = true
	end
	
	--RealisticUtils.testClass("mrGearboxMogli - mrGbMGears", self.mrGbMGears) 
	
	self.mrGbMDefaultGear     = Utils.getNoNil(getXMLInt(xmlFile, xmlString .. ".gears#defaultGear"), 1) 
	self.mrGbMDefaultRange    = Utils.getNoNil(getXMLInt(xmlFile, xmlString .. ".ranges(0)#defaultRange"), table.getn(self.mrGbMRanges)) 
	self.mrGbMDefaultRange2   = Utils.getNoNil(getXMLInt(xmlFile, xmlString .. ".ranges(1)#defaultRange"), table.getn(self.mrGbMRanges2)) 

	self.mrGbMAutoShiftUpTimer   = 0
	self.mrGbMAutoShiftDownTimer = 0
	self.mrGbMCurrentGearSpeed   = 0
	
	self.mrGbMAutoShiftUpRpm     = getXMLFloat(xmlFile, xmlString .. ".gears#autoUpRpm") 
	self.mrGbMAutoShiftDownRpm   = getXMLFloat(xmlFile, xmlString .. ".gears#autoDownRpm") 
	
	self.mrGbMAutoShiftMinMaxFactor = 1
	if self.mrGbMAutoShiftUpRpm  ~= nil and self.mrGbMAutoShiftDownRpm ~= nil then
		self.mrGbMAutoShiftMinMaxFactor = self.mrGbMAutoShiftUpRpm / self.mrGbMAutoShiftDownRpm
	end
	
	mogliBase11.registerState( self, "mrGbMFirstTimeRun",  false )
	mogliBase11.registerState( self, "mrGbMAutomatic",     self.mrGbMAutoShiftUpRpm ~= nil or self.mrGbMAutoShiftDownRpm ~= nil )
	mogliBase11.registerState( self, "mrGbMDrawText",      "off" )
	mogliBase11.registerState( self, "mrGbMIsOn",          false,                  mrGearboxMogli.mrGbMOnSetIsOn )
	mogliBase11.registerState( self, "mrGbMIsOnOff",       self.mrGbMDefaultOn )
	mogliBase11.registerState( self, "mrGbMReverseActive", false,                  mrGearboxMogli.mrGbMOnSetReverse )
	mogliBase11.registerState( self, "mrGbMNeutralActive", self.mrGbMAutoStartStop,mrGearboxMogli.mrGbMOnSetNeutral ) 	
	mogliBase11.registerState( self, "mrGbMCurrentGear",   self.mrGbMDefaultGear,  mrGearboxMogli.mrGbMOnSetGear ) 
	mogliBase11.registerState( self, "mrGbMCurrentRange",  self.mrGbMDefaultRange, mrGearboxMogli.mrGbMOnSetRange )
	mogliBase11.registerState( self, "mrGbMCurrentRange2", self.mrGbMDefaultRange2,mrGearboxMogli.mrGbMOnSetRange2 )
	mogliBase11.registerState( self, "mrGbMWarningText",   "",                     mrGearboxMogli.mrGbMOnSetWarningText )
	mogliBase11.registerState( self, "mrGbMAutoClutch",    true )
	mogliBase11.registerState( self, "mrGbMClutchTime",    0 )
	mogliBase11.registerState( self, "mrGbMManualClutch",  self.mrGbMMaxClutchPercent, mrGearboxMogli.mrGbMOnSetManualClutch, "float32" )
	mogliBase11.registerState( self, "mrGbMWantedMinRpmRatio", 0, nil, "float32" )

	self.mrGbMWarningTimer       = 0	
	self.mrGbMGearShiftingNeeded = 0 
	self.mrGbMGearShiftingTime   = 0 
	self.mrGbMClutchShiftingTime = 0 
	self.mrGbMDoubleClutch       = false
	self.mrGbMLastMinThrottle    = self.mrGbMIdleRpm / self.mrGbMRatedRpm
	self.mrGbMLastMotorLoad      = 0 
	self.mrGbMEngineRevDownTimeMs= 1000
	self.mrGbMEngineRevUpTimeMs  = 1000
	self.mrGbMEngineRpmMinus     = - self.mrGbMRatedRpm / self.mrGbMEngineRevDownTimeMs 
	self.mrGbMEngineRpmPlus      =   self.mrGbMRatedRpm / self.mrGbMEngineRevUpTimeMs 
	self.mrGbMDirtyFlag          = self:getNextDirtyFlag() 
	self.mrGbMCurrentRpm         = self.mrGbMIdleRpm
	self.mrGbMWheelRpm           = 0
	self.mrGbMClutchPercent      = self.mrGbMMinClutchPercent 
	self.mrGbMWantedMinRpmRatio  = 0
	self.mrGbMWantedAcceleration = 0
	self.mrGbMLastMotorTorque    = 0
	self.realClutchPercent       = 0
	self.mrGbMLastSpeed          = 0

	self.mrGbMLastSumDt          = 0 
	self.mrGbMDrawClutch         = 0 
	self.mrGbMLastDrawClutch     = 0 
  self.mrGbMDrawLoad           = 0   
  self.mrGbMLastDrawLoad       = 0   
  self.mrGbMDrawSpeed          = 0 
  self.mrGbMLastDrawSpeed      = 0 
	self.mrGbMDrawFuel           = 0 
	self.mrGbMLastSlipFactor     = 0
	self.mrGbMLastAxisForward    = 0
	
end 

function mrGearboxMogli:update(dt)
	
	if not ( self.mrGbMFirstTimeRun ) then
 		if self.isServer then
			self:mbSetState11( "mrGbMCurrentGear",   mrGearboxMogli.mrGbMGetNewEntry( self, self.mrGbMGears,   self.mrGbMDefaultGear,   "gear" ) )
			self:mbSetState11( "mrGbMCurrentRange",  mrGearboxMogli.mrGbMGetNewEntry( self, self.mrGbMRanges,  self.mrGbMDefaultRange,  "range" ) )
			self:mbSetState11( "mrGbMCurrentRange2", mrGearboxMogli.mrGbMGetNewEntry( self, self.mrGbMRanges2, self.mrGbMDefaultRange2, "range2" ) )
			self:mbSetState11( "mrGbMFirstTimeRun", true )
			self.mrGbMLowBrakeForceScale = self.motor.lowBrakeForceScale
			self.mrGbMPtoMotorRpmRatio   = self.motor.ptoMotorRpmRatio
			self.mrGbMMotorMinRpm        = self.motor.minRpm
			self.mrGbMMotorMaxRpm        = self.motor.maxRpm
			self.mrGbMMaxPossibleRpm     = self.motor.maxRpm
			self.mrGbMMaxClutchTorque    = self.motor.maxClutchTorque
			
			self.mrGbMTorqueCurveBackup = self.motor.torqueCurve
			self.mrGbMTorqueCurve       = AnimCurve:new( self.motor.torqueCurve.interpolator, self.motor.torqueCurve.interpolatorDegree )
			local zeroTorqueRpm = 0.5*self.motor.minRpm
			self.mrGbMTorqueCurve:addKeyframe( {v=0.01, time=0} )
			self.mrGbMTorqueCurve:addKeyframe( {v=0.1, time=zeroTorqueRpm} )
			for _,k in pairs(self.motor.torqueCurve.keyframes) do
				if k.time > zeroTorqueRpm then 
					self.mrGbMTorqueCurve:addKeyframe( k )
				end
			end		
			self.mrGbMTorqueCurve:addKeyframe( {v=0, time=self.mrGbMOverspeedRpm} )
		else
			self.mrGbMFirstTimeRun = true
		end
	end

	if self.isServer then
		local text = ""
		if self.isAITractorActivated or self.isAIThreshing then
			text = "AI"
		elseif not ( self.isMotorStarted ) then
			text = "off"
		elseif self.mrGbMNeutralActive then
			if self.mrGbMReverseActive then
				text = "neutral (R)" 	
			else
				text = "neutral" 	
			end
		elseif self.mrGbMAutomatic then
			if self.mrGbMReverseActive then
				text = "auto (R)" 	
			else
				text = "auto"
			end
		else
			if self.mrGbMReverseActive then
				text = "reverse" 	
			else
				text = "manual" 	
			end
		end
		
		self:mbSetState11( "mrGbMDrawText", text )
	end
	
	if self:getIsActiveForInput(false) and self.mrGbMFirstTimeRun then
		if InputBinding.hasEvent(InputBinding.mrGearboxMogliON_OFF) then
			self:mbSetState11( "mrGbMIsOnOff", not self.mrGbMIsOnOff ) 
		end
			
		if self.mrGbMIsOn then
			
			local driveControlShuttle = false
			if      g_currentMission.driveControl ~= nil 
					and g_currentMission.driveControl.useModules ~= nil
					and g_currentMission.driveControl.useModules.shuttle
					and self.driveControl ~= nil
					and self.driveControl.shuttle ~= nil 
					and self.driveControl.shuttle.isActive then
				--driveControlShuttle = true
				--if     self.driveControl.shuttle.direction > 0 then
				--	self:mbSetState11( "mrGbMReverseActive", false ) 
				--elseif self.driveControl.shuttle.direction < 0 then
				--	self:mbSetState11( "mrGbMReverseActive", true ) 
				--else
				--	self:mbSetState11( "mrGbMNeutralActive", true) 
				--end
				self.driveControl.shuttle.isActive = false
			end
		
			if      self.mrGbMWarningText ~= nil
					and self.mrGbMWarningText ~= "" then
				if g_currentMission.time < self.mrGbMWarningTimer then
					g_currentMission:addWarning(self.mrGbMWarningText, 0.018, 0.033)
				else
					self.mrGbMWarningText = ""
				end
			end
			-- auto start/stop
			if      self.mrGbMAutoStartStop
					and self.mrGbMNeutralActive 			
					and ( self.mrGbMLastAxisForward < -0.1 or self.cruiseControl.state ~= 0 ) then
				self:mbSetState11( "mrGbMNeutralActive", false) 
			end

			if InputBinding.hasEvent(InputBinding.mrGearboxMogliAUTOCLUTCH) then
				self:mbSetState11( "mrGbMAutoClutch", not self.mrGbMAutoClutch ) 
			end

			local targetClutchPercent = InputBinding.getDigitalInputAxis(InputBinding.mrGearboxMogliCLUTCH)
			if InputBinding.isAxisZero(targetClutchPercent) then
				targetClutchPercent = InputBinding.getAnalogInputAxis(InputBinding.mrGearboxMogliCLUTCH)
				if not InputBinding.isAxisZero(targetClutchPercent) then
					targetClutchPercent = Utils.clamp( targetClutchPercent, 0, 1 ) * self.mrGbMMaxClutchPercent
					if math.abs( targetClutchPercent - self.mrGbMManualClutch ) > 0.01 then
						self:mbSetState11( "mrGbMClutchTime", g_currentMission.time )
					end
					self:mbSetState11( "mrGbMManualClutch", targetClutchPercent ) 
				end
			elseif targetClutchPercent < 0 then
				self:mbSetState11( "mrGbMClutchTime", g_currentMission.time )
				self:mbSetState11( "mrGbMManualClutch", math.max( 0, self.mrGbMManualClutch - dt/self.mrGbMClutchTimeDec ))
			elseif targetClutchPercent > 0 then
				self:mbSetState11( "mrGbMClutchTime", g_currentMission.time )
				self:mbSetState11( "mrGbMManualClutch", math.min( self.mrGbMMaxClutchPercent, self.mrGbMManualClutch + dt/self.mrGbMClutchTimeInc ))
			end
			
			local handThrottle = InputBinding.getDigitalInputAxis(InputBinding.mrGearboxMogliMINRPM)
			if InputBinding.isAxisZero(handThrottle) then
				handThrottle = InputBinding.getAnalogInputAxis(InputBinding.mrGearboxMogliMINRPM)
				if not InputBinding.isAxisZero(handThrottle) then
					self:mbSetState11( "mrGbMWantedMinRpmRatio", Utils.clamp( handThrottle, 0, 1 ))
				end
			elseif handThrottle < 0 then
				self:mbSetState11( "mrGbMWantedMinRpmRatio", math.max( 0, self.mrGbMWantedMinRpmRatio - 0.001 * dt )) 
			elseif handThrottle > 0 then
				self:mbSetState11( "mrGbMWantedMinRpmRatio", math.min( 1, self.mrGbMWantedMinRpmRatio + 0.001 * dt )) 
			end
				
			if     InputBinding.hasEvent(InputBinding.mrGearboxMogliNEUTRAL) then
				if self.mrGbMAutoShiftUpRpm ~= nil or self.mrGbMAutoShiftDownRpm ~= nil then
					if self.mrGbMNeutralActive == self.mrGbMAutomatic then
						self:mbSetState11( "mrGbMNeutralActive", not self.mrGbMNeutralActive ) 
						self:mrGbMSetAutomatic( true ) 
					else
						self:mrGbMSetAutomatic( not self.mrGbMAutomatic ) 
					end					
				else	
					if not self.mrGbMNeutralActive then
						self:setCruiseControlState(0)
					end
					self:mbSetState11( "mrGbMNeutralActive", not self.mrGbMNeutralActive ) 
				end
			elseif not ( driveControlShuttle ) and InputBinding.hasEvent(InputBinding.mrGearboxMogliREVERSE) then			
				self:setCruiseControlState(0)
				self:mbSetState11( "mrGbMReverseActive", not self.mrGbMReverseActive ) 
			elseif InputBinding.hasEvent(InputBinding.mrGearboxMogliSHIFTRANGE2UP) then -- high/low range shift
				self:mrGbMSetRange2(self.mrGbMCurrentRange2+1)                                        
			elseif InputBinding.hasEvent(InputBinding.mrGearboxMogliSHIFTRANGE2DOWN) then -- high/low range shift
				self:mrGbMSetRange2(self.mrGbMCurrentRange2-1) 
			elseif InputBinding.hasEvent(InputBinding.mrGearboxMogliSHIFTRANGEUP) then -- high/low range shift
				if self.mrGbMSwapGearRangeKeys then
					self:mrGbMSetGear(self.mrGbMCurrentGear+1) 
				else
					self:mrGbMSetRange(self.mrGbMCurrentRange+1)                                        
				end 
			elseif InputBinding.hasEvent(InputBinding.mrGearboxMogliSHIFTRANGEDOWN) then -- high/low range shift
				if self.mrGbMSwapGearRangeKeys then
					self:mrGbMSetGear(self.mrGbMCurrentGear-1) 	
				else
					self:mrGbMSetRange(self.mrGbMCurrentRange-1) 
				end 
			elseif InputBinding.hasEvent(InputBinding.mrGearboxMogliSHIFTGEARUP) then
				if self.mrGbMSwapGearRangeKeys then
					self:mrGbMSetRange(self.mrGbMCurrentRange+1)                                        
				else
					self:mrGbMSetGear(self.mrGbMCurrentGear+1) 
				end 
			elseif InputBinding.hasEvent(InputBinding.mrGearboxMogliSHIFTGEARDOWN) then
				if self.mrGbMSwapGearRangeKeys then
					self:mrGbMSetRange(self.mrGbMCurrentRange-1) 
				else
					self:mrGbMSetGear(self.mrGbMCurrentGear-1) 	
				end 
			elseif InputBinding.hasEvent(InputBinding.mrGearboxMogliGEAR1) then if self.mrGbMSwapGearRangeKeys then self:mrGbMSetRange(1) else self:mrGbMSetGear(1) end 
			elseif InputBinding.hasEvent(InputBinding.mrGearboxMogliGEAR2) then if self.mrGbMSwapGearRangeKeys then self:mrGbMSetRange(2) else self:mrGbMSetGear(2) end 
			elseif InputBinding.hasEvent(InputBinding.mrGearboxMogliGEAR3) then if self.mrGbMSwapGearRangeKeys then self:mrGbMSetRange(3) else self:mrGbMSetGear(3) end 
			elseif InputBinding.hasEvent(InputBinding.mrGearboxMogliGEAR4) then if self.mrGbMSwapGearRangeKeys then self:mrGbMSetRange(4) else self:mrGbMSetGear(4) end 
			elseif InputBinding.hasEvent(InputBinding.mrGearboxMogliGEAR5) then if self.mrGbMSwapGearRangeKeys then self:mrGbMSetRange(5) else self:mrGbMSetGear(5) end 
			elseif InputBinding.hasEvent(InputBinding.mrGearboxMogliGEAR6) then if self.mrGbMSwapGearRangeKeys then self:mrGbMSetRange(6) else self:mrGbMSetGear(6) end 
			elseif InputBinding.hasEvent(InputBinding.mrGearboxMogliGEARFWD)  then self:setCruiseControlState(0) self:mbSetState11( "mrGbMReverseActive", false ) 
			elseif InputBinding.hasEvent(InputBinding.mrGearboxMogliGEARBACK) then self:setCruiseControlState(0) self:mbSetState11( "mrGbMReverseActive", true ) 
			end 
		end 
	end 
		
--**********************************************************************************************************		
--**********************************************************************************************************		
	if self.isServer and self.isActive and self.mrGbMIsOn and self.mrGbMFirstTimeRun then	
--**********************************************************************************************************		
						
--**********************************************************************************************************		
		-- defaults
		self.motor.torqueCurve       = self.mrGbMTorqueCurve
		self.motor.ptoMotorRpmRatio  = self.mrGbMPtoMotorRpmRatio
		self.mrGbMWantedAcceleration = math.max( -self.mrGbMLastAxisForward, self.mrGbMWantedMinRpmRatio )
		self.mrGbMLastSpeed          = 3600 * self.lastSpeed
		
--**********************************************************************************************************	
		-- current RPM and power
		local minRpm = self.mrGbMIdleRpm-- + self.mrGbMWantedMinRpmRatio * ( self.mrGbMRatedRpm - self.mrGbMIdleRpm - 100 )
		
		if self.attachedImplements ~= nil and self.mrGbMPtoMotorRpmRatio ~= nil then
			local ptoRpm 
			for _,a in pairs(self.attachedImplements) do
				if a.object.getPtoRpm ~= nil and a.object.getDoConsumePtoPower ~= nil and a.object:getDoConsumePtoPower() then
					local r = a.object:getPtoRpm( )
					if r ~= nil and ( ptoRpm == nil or ptoRpm < r ) then
						ptoRpm = r
					end
				end
			end
			
			if ptoRpm ~= nil then
				local p = math.min( ptoRpm * self.mrGbMPtoMotorRpmRatio, self.mrGbMRatedRpm - 300 )
				self.motor.ptoMotorRpmRatio = ( p - 50 ) / ptoRpm
				if minRpm < p then
					minRpm = p
				end
			end
		end
		
		self.motor.minRpm        = self.mrGbMIdleRpm - 100		
		
		local currentWheelRpm    = self.motor.clutchRpm --math.max( self.lastSpeed * 1000 / math.max( self.mrGbMCurrentGearSpeed, 1E-3 ), 0 ) * self.mrGbMRatedRpm
		local currentEngineRpm   = self.motor.nonClampedMotorRpm
		
		self.mrGbMWheelRpm       = self.mrGbMWheelRpm   + 0.1 * ( currentWheelRpm  - self.mrGbMWheelRpm )
		self.mrGbMCurrentRpm     = self.mrGbMCurrentRpm + 0.1 * ( currentEngineRpm - self.mrGbMCurrentRpm )
		self.mrGbMNoTransmission = false
		self.mrGbMMaxPossibleRpm = self.mrGbMMotorMaxRpm * self.motor:getBestGearRatio() / self.motor.minForwardGearRatio 

		--currentWheelRpm          = self.mrGbMWheelRpm
		--currentEngineRpm         = self.mrGbMCurrentRpm 
		local wrongDirection     = false
		
		local motorLoad = math.sqrt( Utils.clamp( self.motor.motorLoad, 0, 1 ) )
		self.mrGbMLastMotorLoad	= self.mrGbMLastMotorLoad	+ 0.04 * ( motorLoad - self.mrGbMLastMotorLoad )
		
--**********************************************************************************************************		
		if     self.mrGbMNeutralActive
		    or ( currentEngineRpm             <= self.mrGbMIdleRpm + 100
		     and self.mrGbMLastAxisForward    >  0.01
				 and self.cruiseControl.state     == 0 ) then
		-- neutral or braking
		  if self.mrGbMLastSpeed < 0.3 and self.mrGbMAutoStartStop then
				self:mbSetState11( "mrGbMNeutralActive", true) 
			end
			
			if self.mrGbMGearShiftingNeeded > 0 then
				if g_currentMission.time>=self.mrGbMGearShiftingTime then
					if self.mrGbMGearShiftingNeeded < 2 then	
						self:mrGbMDoGearShift() 
					end 
					self.mrGbMGearShiftingNeeded = 0 
				elseif self.mrGbMGearShiftingNeeded < 2 then	
					if self.realClutchPercent <= 0 then
						self:mrGbMDoGearShift() 
					end 
				end 
			end

			if self.mrGbMAutoClutch then
				self.realClutchPercent  = math.max( 0, self.realClutchPercent -dt/self.mrGbMClutchTimeDec ) 
			end
			self.mrGbMLastSlipFactor  = 0
		  self.mrGbMNoTransmission  = true
			self.mrGbMLastMinThrottle = self.mrGbMWantedMinRpmRatio
--**********************************************************************************************************		
		else
--**********************************************************************************************************		
			-- acceleration for idle/minimum rpm
			local delta     = currentEngineRpm - minRpm 
			local throttle  = self.mrGbMLastMotorLoad * ( minRpm - 0.7 * self.mrGbMIdleRpm ) / ( self.mrGbMRatedRpm - 0.7 * self.mrGbMIdleRpm ) - Utils.clamp( 0.005 * delta, -0.1, 0.1 )
			local throttle  = Utils.clamp( throttle, self.mrGbMWantedMinRpmRatio, 1 ) 
			self.mrGbMLastMinThrottle = throttle -- self.mrGbMLastMinThrottle + 0.3 * ( throttle - self.mrGbMLastMinThrottle )
						
			if self.mrGbMGearShiftingNeeded > 0 then
			-- during gear shifting
				
				if not ( self.mrGbMAutoClutch ) then
				elseif self.mrGbMGearShiftingNeeded == 2 and g_currentMission.time < self.mrGbMGearShiftingTime then	
					if currentEngineRpm > 0.9 * self.mrGbMRatedRpm then
						self.mrGbMGearShiftingNeeded  = 3
					else
						self.mrGbMNoTransmission = true
					end
					self.mrGbMCurrentRpm       = Utils.clamp( self.mrGbMCurrentRpm + dt*self.mrGbMRatedRpm/self.mrGbMEngineRevUpTimeMs, minRpm, self.mrGbMMotorMaxRpm ) 
				else
					self.mrGbMCurrentRpm       = Utils.clamp( self.mrGbMCurrentRpm - dt*self.mrGbMRatedRpm/self.mrGbMEngineRevDownTimeMs, minRpm, self.mrGbMMotorMaxRpm )
				end

				if g_currentMission.time >= self.mrGbMGearShiftingTime then
					if self.mrGbMGearShiftingNeeded < 2 then	
						self:mrGbMDoGearShift() 
					end 
					self.mrGbMGearShiftingNeeded = 0 
					if self.mrGbMAutoClutch then
						self.realClutchPercent     = self.mrGbMClutchPercent 
					end
					if currentWheelRpm > self.mrGbMWheelRpm then
						currentWheelRpm = self.mrGbMWheelRpm
					end
				elseif self.mrGbMGearShiftingNeeded < 2 then	
					if self.realClutchPercent > 0 and g_currentMission.time < self.mrGbMClutchShiftingTime then
						if self.mrGbMAutoClutch then
							self.realClutchPercent   = Utils.clamp( ( self.mrGbMClutchShiftingTime - g_currentMission.time )/self.mrGbMClutchShiftTime, 0, self.realClutchPercent ) 					
						end
					else
						--self.realClutchPercent     = 0 
						self:mrGbMDoGearShift() 
					end 
				else
					--self.realClutchPercent       = 0 
				end 
				
				if self.mrGbMAutoClutch then
					self.mrGbMMaxPossibleRpm  = self.mrGbMCurrentRpm
				end
				self.mrGbMNoTransmission  = true
				self.mrGbMLastSlipFactor  = 0
				
	----**********************************************************************************************************		
	--		elseif  self.mrGbMLastSpeed > 0.3
	--				and ( (     self.mrGbMReverseActive and self.movingDirection > 0 ) 
	--					 or ( not self.mrGbMReverseActive and self.movingDirection < 0 ) ) then
	--		-- wrong direction
	--			self.mrGbMCurrentRpm      = Utils.clamp( self.mrGbMCurrentRpm - dt*self.mrGbMRatedRpm/self.mrGbMEngineRevDownTimeMs, minRpm, self.mrGbMMotorMaxRpm )
	--			self.mrGbMLastSlipFactor  = 0
	--			self.mrGbMNoTransmission  = true
	--			wrongDirection            = true
			else
	--**********************************************************************************************************		
	-- normal drive with gear and clutch
				self.mrGbMNoTransmission = false
				self.motor.lowBrakeForceScale = self.mrGbMRealMotorBrakeFx

	--**********************************************************************************************************		
	-- automatic shifting
				--no auto parkbrake when a gear is engaged
				local factorFx     = self.mrGbMLastMotorLoad 
				
				-- semi automatic transmission
				if self.mrGbMAutomatic then
					local diff = 0 
					local minRpmReduced = math.max( self.mrGbMIdleRpm, minRpm - 100 )
					local autoDownRpm, autoUpRpm = minRpmReduced, self.mrGbMRatedRpm
					
					if self.mrGbMAutoShiftDownRpm ~= nil then
						if      factorFx                   <  1
								and self.mrGbMAutoShiftDownRpm >  minRpmReduced then
							diff = (1 - factorFx) * ( self.mrGbMAutoShiftDownRpm - minRpmReduced ) 
						else
							diff = 0
						end
						autoDownRpm = math.max( minRpmReduced, self.mrGbMAutoShiftDownRpm - diff ) 
					end 
					if self.mrGbMAutoShiftUpRpm ~= nil then
						if      factorFx                   <  1
								and self.mrGbMAutoShiftDownRpm ~= 0 
								and self.mrGbMAutoShiftUpRpm   >  minRpmReduced * self.mrGbMAutoShiftMinMaxFactor then
							diff = (1 - factorFx) * ( self.mrGbMAutoShiftUpRpm - minRpmReduced * self.mrGbMAutoShiftMinMaxFactor ) 
						else
							diff = 0
						end
						autoUpRpm  = math.min( self.mrGbMRatedRpm, self.mrGbMAutoShiftUpRpm - diff ) 
					end 

					--print(string.format("%4.0f %4.0f %4.0f %4.0f %4.0f",currentWheelRpm,currentEngineRpm,minRpmReduced,autoDownRpm,autoUpRpm))
																			
					local wheelRpm
					
					-- shift down			
					wheelRpm = math.max( currentWheelRpm, self.mrGbMWheelRpm )
					if 			self.mrGbMAutoShiftDownRpm ~= nil 
							and self.mrGbMCurrentGear      > 1 
							and ( wheelRpm                 < minRpmReduced
								 or ( g_currentMission.time  > self.mrGbMAutoShiftDownTimer 
									and self.mrGbMWantedAcceleration > 0.1 
									and wheelRpm               < autoDownRpm )) then
						self:mrGbMSetGear(self.mrGbMCurrentGear-1) 
					
					-- shift up				
					wheelRpm = math.min( currentWheelRpm, self.mrGbMWheelRpm )
					elseif  self.mrGbMAutoShiftUpRpm   ~= nil 
							and self.mrGbMCurrentGear      < table.getn( self.mrGbMGears )
							and ( wheelRpm                 > self.mrGbMRatedRpm
								 or ( g_currentMission.time  > self.mrGbMAutoShiftUpTimer 	
									and self.mrGbMWantedAcceleration > 0.1 
									and wheelRpm               > autoUpRpm )) then
						self:mrGbMSetGear(self.mrGbMCurrentGear+1) 
					end
				end
			

	--**********************************************************************************************************		
	-- clutch
				local slipFactor = 1
				if self.motor.nonClampedMotorRpm < self.mrGbMCloseRpm then
					slipFactor = Utils.clamp( self.motor.nonClampedMotorRpm / self.motor.lastMotorRpm, 0, 1 )
				end
				self.mrGbMLastSlipFactor = self.mrGbMLastSlipFactor + 0.1 * ( slipFactor - self.mrGbMLastSlipFactor )
				
				local openRpm = minRpm
				if     self.mrGbMOpenRpm <= openRpm then
					openRpm = self.mrGbMOpenRpm
				elseif self.mrGbMOpenRpm > minRpm then
					openRpm = minRpm + factorFx * ( self.mrGbMOpenRpm - minRpm )
				end
					
				local refRpm   = math.max( minRpm, currentEngineRpm )
				local closeRpm = refRpm
				if self.mrGbMCloseRpm > refRpm then
					closeRpm = refRpm + factorFx * ( self.mrGbMCloseRpm - refRpm )
				end
				
				if self.mrGbMAutoClutch then
					local targetClutchPercent = mrGearboxMogli.mrGbMGetClutchPercent( self, math.min( self.mrGbMCurrentRpm, currentEngineRpm ), openRpm, closeRpm )
					if      dt > self.mrGbMClutchTimeDec
							and dt > self.mrGbMClutchTimeInc then
						self.realClutchPercent = Utils.clamp( targetClutchPercent, 0, self.mrGbMMaxClutchPercent ) 
					elseif targetClutchPercent < self.realClutchPercent then
						self.realClutchPercent = math.max( 0, self.realClutchPercent - math.min( dt/self.mrGbMClutchTimeDec, self.realClutchPercent - targetClutchPercent ) ) 
					else
						self.realClutchPercent = math.min( self.mrGbMMaxClutchPercent, self.realClutchPercent + math.min( dt/self.mrGbMClutchTimeInc, targetClutchPercent - self.realClutchPercent ) ) 
					end 
				else
					self.realClutchPercent   = self.mrGbMMaxClutchPercent
				end 					
			end 					
		end
		
		if self.mrGbMAutoClutch then
			if self.mrGbMClutchTime <= g_currentMission.time and g_currentMission.time < self.mrGbMClutchTime + 5000 then
				clutchPercent = self.mrGbMManualClutch + ( g_currentMission.time - self.mrGbMClutchTime  ) / 5000 * ( self.realClutchPercent - self.mrGbMManualClutch  )
			else
				clutchPercent = self.realClutchPercent
			end
		else
			clutchPercent = self.mrGbMManualClutch
		end
		
		if self.mrGbMNoTransmission or clutchPercent <= 0 then
			self.motor.maxClutchTorque = 0
			self.motor.minRpm          = self.mrGbMIdleRpm
		else
			local c = clutchPercent
			if self.mrGbMClutchExponent < 0.999 or self.mrGbMClutchExponent > 1.001 then	
				c = c ^ self.mrGbMClutchExponent
			end
			
		--local mct = c * ( self.mrGbMClutchTorqueConst + self.mrGbMClutchTorqueDiff * math.max( 1 - currentWheelRpm / math.min( currentEngineRpm, self.mrGbMMaxPossibleRpm ), 0 ) )--self.mrGbMTorqueCurve:get( self.motor.nonClampedMotorRpm ) * 
			local mct = c * ( self.mrGbMClutchTorqueConst + self.mrGbMClutchTorqueDiff * Utils.clamp( ( currentEngineRpm - currentWheelRpm ) / self.mrGbMMotorMaxRpm, 0, 1 )^2 )-- * math.max( 1 - currentWheelRpm / math.min( currentEngineRpm, self.mrGbMMaxPossibleRpm ), 0 ) )

			local f1 = 0.9
			local f2 = 1 - f1
			if clutchPercent > f1 then
				mct = math.max( self.mrGbMClutchTorqueClosed * ( clutchPercent - f1 ) / f2, mct )
		--elseif currentWheelRpm < self.mrGbMCloseRpm then
		--	self.mrGbMMaxPossibleRpm   = math.min( self.motor.maxRpm - 1.111 * clutchPercent * math.max( self.motor.maxRpm - self.mrGbMCloseRpm, 0 ), self.mrGbMMaxPossibleRpm )
			end

			self.mrGbMMaxPossibleRpm   = (1-clutchPercent) * self.mrGbMMotorMaxRpm + clutchPercent * currentWheelRpm
			self.mrGbMMaxPossibleRpm   = self.mrGbMMaxPossibleRpm + dt * self.mrGbMRpmIncFactor
			self.motor.maxClutchTorque = mct --self.motor.maxClutchTorque + 0.1 * ( mct - self.motor.maxClutchTorque )							
			self.motor.maxRpm          = self.mrGbMOverspeedRpm
		end
		
		--if self.mrGbMLastAxisForward > 0 then
		--	self.brakeLightsVisibility = true
		--else
		--	self.brakeLightsVisibility = false
		--end
		--self.reverseLightsVisibility = self.mrGbMReverseActive
	end 	
--**********************************************************************************************************		
--**********************************************************************************************************		

end 

function mrGearboxMogli:mrGbMGetClutchPercent( currrentRpm, openRpm, closeRpm, gearSpeed )
	local c = self.realClutchPercent
	if currrentRpm <= self.motor.minRpm and openRpm > 0 then
		c = self.mrGbMMinClutchPercent
	elseif currrentRpm >= closeRpm then
		c = self.mrGbMMaxClutchPercent 					
	elseif closeRpm - openRpm >= 10 and currrentRpm >= openRpm then
		c = self.realClutchPercent + ( self.mrGbMMaxClutchPercent - self.realClutchPercent ) * ( currrentRpm - openRpm ) / ( closeRpm - openRpm )
	elseif openRpm - self.motor.minRpm >= 10 then
		c = self.mrGbMMinClutchPercent + ( self.realClutchPercent - self.mrGbMMinClutchPercent ) * ( currrentRpm - self.motor.minRpm ) / ( openRpm - self.motor.minRpm )
	elseif currrentRpm <= openRpm then
		c = self.mrGbMMinClutchPercent
	end
	return c	
end

function mrGearboxMogli:mrGbMDoGearShift(increaseOnly)
	if self.isServer then
		if self.mrGbMDoubleClutch then
			self.mrGbMGearShiftingNeeded = 2
		else
			self.mrGbMGearShiftingNeeded = 3
		end
		local gearMaxSpeed = self.mrGbMGears[self.mrGbMCurrentGear].speed 
		                   * self.mrGbMRanges[self.mrGbMCurrentRange].ratio 
											 * self.mrGbMRanges2[self.mrGbMCurrentRange2].ratio
											 * self.mrGbMGlobalRatioFactor
		if self.mrGbMReverseActive then	
			gearMaxSpeed = gearMaxSpeed * self.mrGbMReverseRatio 
		end
		
		--mogliBase11.debugEvent( self, gearMaxSpeed, self.mrGbMCurrentGearSpeed, increaseOnly )
		
		if increaseOnly then
			self.mrGbMCurrentGearSpeed = math.max( gearMaxSpeed, self.mrGbMCurrentGearSpeed )
		else
			self.mrGbMCurrentGearSpeed = gearMaxSpeed 
		end
		
		self.mrSEFLastEngineRevFx = nil
	end 
end 

function mrGearboxMogli:updateTick(dt)

	if self.isActive then
		if self.isServer then
			if self.isAITractorActivated or self.isAIThreshing then
				if self.mrGbMIsOn then
					self:mbSetState11( "mrGbMNeutralActive", true )
					self.motor.rpmLimit            = self.motor.maxRpm
				end
				self:mbSetState11( "mrGbMIsOn", false ) 				
			elseif  self.isMotorStarted 
					and self.mrGbMIsOnOff
					and self.mrGbMFirstTimeRun then
				self:mbSetState11( "mrGbMIsOn", true ) 
			else		
				self:mbSetState11( "mrGbMIsOn", false ) 
			end 	
		
			if self.mrGbMIsOn then	
					self.mrGbMLastSumDt = self.mrGbMLastSumDt + dt
				
				if self.mrGbMLastSumDt > 200 then
					self.mrGbMDrawClutch = tonumber( math.floor(math.min( self.realClutchPercent, self.mrGbMManualClutch ) * self.mrGbMMaxClutchPercentInv*200+0.5) ) 
					self.mrGbMDrawLoad   = tonumber( math.floor(Utils.getNoNil(self.mrGbMLastMotorLoad,0)*20+0.5)*5 ) 			
					self.mrGbMDrawSpeed  = tonumber( math.floor(self.mrGbMCurrentGearSpeed / self.mrGbMGlobalRatioFactor * 3.6 + 0.5 ) ) 

					if self.mrGbMLastFuelFillLevel == nil then
						self.mrGbMLastFuelFillLevel = self.fuelFillLevel
						self.mrGbMDrawFuel          = 0
					else
						local fuelUsed = self.mrGbMLastFuelFillLevel - self.fuelFillLevel
						self.mrGbMLastFuelFillLevel = self.fuelFillLevel
						self.mrGbMDrawFuel          = fuelUsed * (1000 * 3600) / self.mrGbMLastSumDt
					end
					
					if     self.mrGbMLastDrawClutch ~= self.mrGbMDrawClutch
							or self.mrGbMLastDrawLoad   ~= self.mrGbMDrawLoad   
							or self.mrGbMLastDrawSpeed  ~= self.mrGbMDrawSpeed 
							then
						self:raiseDirtyFlags(self.mrGbMDirtyFlag) 
					
						self.mrGbMLastDrawClutch = self.mrGbMDrawClutch
						self.mrGbMLastDrawLoad   = self.mrGbMDrawLoad   
						self.mrGbMLastDrawSpeed  = self.mrGbMDrawSpeed 
					end 
					
					self.mrGbMLastSumDt = 0
				end 
			else
				self.mrGbMLastSumDt         = 0
				self.mrGbMLastFuelFillLevel = self.fuelFillLevel
			end
		end
	end
end 

function mrGearboxMogli:readUpdateStream(streamId, timestamp, connection)
  if connection:getIsServer() then
		if streamReadBool( streamId ) then
			self.mrGbMDrawClutch = streamReadUIntN( streamId ) 
			self.mrGbMDrawLoad   = streamReadUIntN( streamId )  
			self.mrGbMDrawSpeed  = streamReadUIntN( streamId ) 
		end 
  end 
end 

function mrGearboxMogli:writeUpdateStream(streamId, connection, dirtyMask)
  if not connection:getIsServer() then
		if streamWriteBool(streamId, bitAND(dirtyMask, self.mrGbMDirtyFlag) ~= 0) then	
			streamWriteUIntN(streamId, self.mrGbMDrawClutch ) 
			streamWriteUIntN(streamId, self.mrGbMDrawLoad   ) 
			streamWriteUIntN(streamId, self.mrGbMDrawSpeed  ) 
		end 
	end 
end 

function mrGearboxMogli:deleteMap()
	if mrGearboxMogli.backgroundOverlayId ~= nil then
		pcall( delete, mrGearboxMogli.backgroundOverlayId )
		mrGearboxMogli.backgroundOverlayId = nil
	end
end

function mrGearboxMogli:draw() 	
	
	if self.mrGbMIsOn then
		if mrGearboxMogli.backgroundOverlayId == nil then
			mrGearboxMogli.backgroundOverlayId = createImageOverlay( "dataS2/menu/blank.png" )
			setOverlayColor( mrGearboxMogli.backgroundOverlayId, 0,0,0, 0.4 )
		end	
		
		local drawY0   = 0.706
		local deltaY   = 0.02

		local ovTop    = drawY0 + 0.06  -- title plus 0.03 plus character size of title (0.03)
		local ovRows   = 6
		if self.mrGbMDrawSpeed > 0 then
			ovRows = ovRows + 1
		end
		local ovH      = 0.03 + 0.02 + ovRows * deltaY -- title is 0.03 points above drawY0; add border of 0.01 x 2
		local ovY      = ovTop - ovH
		
		renderOverlay( mrGearboxMogli.backgroundOverlayId, 0.84, ovY, 0.16, ovH )
	
		setTextAlignment(RenderText.ALIGN_LEFT) 
		setTextColor(1, 1, 1, 1) 

		setTextBold(true) 
		renderText(0.85, drawY0 + 0.03, 0.03, self.mrGbMDrawText) 			     	
		setTextBold(false) 
		
		local gearText = self.mrGbMGears[self.mrGbMCurrentGear].name
		if self.mrGbMRanges[self.mrGbMCurrentRange].name ~= nil and self.mrGbMRanges[self.mrGbMCurrentRange].name ~= "" then
			if self.mrGbMSwapGearRangeKeys then
				gearText = gearText .." ".. self.mrGbMRanges[self.mrGbMCurrentRange].name
			else
				gearText = self.mrGbMRanges[self.mrGbMCurrentRange].name .." ".. gearText
			end
		end
		if self.mrGbMRanges2[self.mrGbMCurrentRange2].name ~= nil and self.mrGbMRanges2[self.mrGbMCurrentRange2].name ~= "" then
			gearText = self.mrGbMRanges2[self.mrGbMCurrentRange2].name .." ".. gearText
		end
		
		local drawY    = drawY0 
		
		renderText(0.85, drawY, deltaY, gearText) 			     
		if self.mrGbMDrawSpeed > 0 then
			drawY = drawY - deltaY renderText(0.85, drawY, deltaY, "Speed.................")
		end
		drawY = drawY - deltaY renderText(0.85, drawY, deltaY, "Rpm...................")
		drawY = drawY - deltaY renderText(0.85, drawY, deltaY, "Load..................")
		drawY = drawY - deltaY renderText(0.85, drawY, deltaY, "Fuel used.............")
		if self.mrGbMAutoClutch then 
			drawY = drawY - deltaY renderText(0.85, drawY, deltaY, "Auto clutch...........")
		else
			drawY = drawY - deltaY renderText(0.85, drawY, deltaY, "Clutch................")
		end
		drawY = drawY - deltaY renderText(0.85, drawY, deltaY, "Hand throttle.........")	
		setTextAlignment(RenderText.ALIGN_RIGHT) 
		
		drawY    = drawY0 
		
		if self.mrGbMDrawSpeed > 0 then
			drawY = drawY - deltaY renderText(0.99,drawY,deltaY, string.format("%3.1f km/h", self.mrGbMDrawSpeed ))
		end
		drawY = drawY - deltaY renderText(0.99, drawY, deltaY, string.format("%4.0f", math.floor(self.motor.lastMotorRpm*0.1+0.5)*10)) 		
		drawY = drawY - deltaY renderText(0.99, drawY, deltaY, string.format("%3d %%", self.mrGbMDrawLoad )) 	
		drawY = drawY - deltaY renderText(0.99, drawY, deltaY, string.format("%3d l/h", self.mrGbMDrawFuel ))  		          
		if self.mrGbMAutoClutch then 
			drawY = drawY - deltaY renderText(0.99, drawY, deltaY, string.format("%3.0f %%", self.mrGbMDrawClutch*0.5 )) 
		else
			drawY = drawY - deltaY renderText(0.99, drawY, deltaY, string.format("%3.0f %%", math.floor( self.mrGbMManualClutch * self.mrGbMMaxClutchPercentInv * 100 + 0.5 ) ))
		end
		drawY = drawY - deltaY renderText(0.99, drawY, deltaY, string.format("%3d %%", math.floor( self.mrGbMWantedMinRpmRatio * 100 + 0.5 ) ))  		          

		setTextAlignment(RenderText.ALIGN_LEFT) 
		setTextBold(false)
		
		g_currentMission:addHelpButtonText(mogliBase11.getText("mrGearboxMogliON"),  InputBinding.mrGearboxMogliON_OFF);		
	else
		g_currentMission:addHelpButtonText(mogliBase11.getText("mrGearboxMogliOFF"), InputBinding.mrGearboxMogliON_OFF);		
	end
	
end 

function mrGearboxMogli:getSaveAttributesAndNodes(nodeIdent)

	local attributes = ""

	attributes = attributes.." mrGbMCurrentGear=\""  .. tostring(self.mrGbMCurrentGear  ) .. "\""
	attributes = attributes.." mrGbMCurrentRange=\"" .. tostring(self.mrGbMCurrentRange ) .. "\""
	attributes = attributes.." mrGbMCurrentRange2=\"" ..tostring(self.mrGbMCurrentRange2 ) .. "\""

	return attributes
end 

function mrGearboxMogli:loadFromAttributesAndNodes(xmlFile, key, resetVehicles)
	local i
	
	i = getXMLInt(xmlFile, key .. "#mrGbMCurrentGear" )
	if i ~= nil then
		self.mrGbMDefaultGear = i
	end

	i = getXMLInt(xmlFile, key .. "#mrGbMCurrentRange" )
	if i ~= nil then
		self.mrGbMDefaultRange = i
	end
	
	i = getXMLInt(xmlFile, key .. "#mrGbMCurrentRange2" )
	if i ~= nil then
		self.mrGbMDefaultRange2 = i
	end

	self:mbSetState11( "mrGbMFirstTimeRun", false )
	
	return BaseMission.VEHICLE_LOAD_OK
end 

function mrGearboxMogli:mrGbMGetNewEntry( entries, index, name )

	local new = Utils.clamp( index, 1, table.getn( entries ) )

	if self.mrGbMReverseActive then
		if new > self.mrGbMCurrentGear then
			while entries[new].forwardOnly and new < table.getn( entries ) do
				new = new + 1
			end
		end
		while entries[new].forwardOnly and new > 1 do
			new = new -1
		end
		if entries[new].forwardOnly then
			while entries[new].forwardOnly and new < table.getn( entries ) do
				new = new + 1
			end
		end
		if entries[new].forwardOnly then
			print(string.format("no reverse %s found: %d", index, name))
		end
	else
		if new > self.mrGbMCurrentGear then
			while entries[new].reverseOnly and new < table.getn( entries ) do
				new = new + 1
			end
		end
		while entries[new].reverseOnly and new > 1 do
			new = new -1
		end
		if entries[new].reverseOnly then
			while entries[new].reverseOnly and new < table.getn( entries ) do
				new = new + 1
			end
		end
		if entries[new].reverseOnly then
			print(string.format("no forward %s found: %d", index, name))
		end
	end
	
	return new
end

function mrGearboxMogli:mrGbMSpeedLevelCheck( newGear, newRange, newRange2 )
	--if 0 < self.motor.speedLevel and not self.mrGbMAutomatic then
	--	local maxGearSpeed = 3.6 * self.mrGbMGears[newGear].speed * self.mrGbMRanges[newRange].ratio * self.mrGbMRanges2[newRange2].ratio
	--	if self.motor.speedLevel == 4 then
	--		maxGearSpeed = maxGearSpeed * self.mrGbMOverspeedRpmRatio
	--	else
	--		maxGearSpeed = maxGearSpeed * self.mrGbMGlobalRatioFactorx
	--	end
  --
	--	if self.mrGbMReverseActive then	
	--		maxGearSpeed = maxGearSpeed * self.mrGbMReverseRatio 
	--	end
	--	if maxGearSpeed < self.motor.realSpeedLevels[self.motor.speedLevel] then
	--		local gearText = self.mrGbMGears[newGear].name
	--		if self.mrGbMRanges[newRange].name ~= nil and self.mrGbMRanges[newRange].name ~= "" then
	--			gearText = self.mrGbMRanges[newRange].name .." ".. gearText
	--		end           
	--		if self.mrGbMRanges2[newRange2].name ~= nil and self.mrGbMRanges2[newRange2].name ~= "" then
	--			gearText = self.mrGbMRanges2[newRange2].name .." ".. gearText
	--		end
	--	
	--		self:mbSetState11( "mrGbMWarningText", string.format( "Gear %s is too low for speed level %d", gearText, self.motor.speedLevel ))
  --
	--		return false
	--	end
	--end
	--
	return true
end

function mrGearboxMogli:mrGbMSetGear( new, noEventSend)
	if      not ( self.mrGbMAutoClutch )
			and self.mrGbMManualClutch > self.mrGbMClutchAfterShiftGear + 0.1 
			and self.mrGbMManualClutch > self.mrGbMClutchPercent then
		return
	end		

	local newGear  = mrGearboxMogli.mrGbMGetNewEntry( self, self.mrGbMGears, new, "gear" )
	local newRange = self.mrGbMCurrentRange
	if not self.mrGbMAutomatic then
		if     newGear > self.mrGbMCurrentGear then 
			newRange = self.mrGbMCurrentRange + self.mrGbMGears[self.mrGbMCurrentGear].upRangeOffset
		elseif newGear < self.mrGbMCurrentGear then 
			newRange = self.mrGbMCurrentRange + self.mrGbMGears[self.mrGbMCurrentGear].downRangeOffset
		end
		newRange = mrGearboxMogli.mrGbMGetNewEntry( self, self.mrGbMRanges, newRange, "range" )
	end
	
	if      newGear ~= self.mrGbMCurrentGear 
			and mrGearboxMogli.mrGbMSpeedLevelCheck( self, newGear, newRange, self.mrGbMCurrentRange2 ) then
		self:mbSetState11( "mrGbMCurrentGear",  newGear,  noEventSend ) 		
		self:mbSetState11( "mrGbMCurrentRange", newRange, noEventSend ) 
	end
end 

function mrGearboxMogli:mrGbMSetRange(new, noEventSend)
	if      not ( self.mrGbMAutoClutch )
			and self.mrGbMManualClutch > self.mrGbMClutchAfterShiftHl + 0.1 
			and self.mrGbMManualClutch > self.mrGbMClutchPercent then
		return
	end		

	local newRange = mrGearboxMogli.mrGbMGetNewEntry( self, self.mrGbMRanges, new, "range" )
	local newGear  = self.mrGbMCurrentGear
	if     newRange > self.mrGbMCurrentRange then 
		newGear = self.mrGbMCurrentGear + self.mrGbMRanges[self.mrGbMCurrentRange].upGearOffset
	elseif newRange < self.mrGbMCurrentRange then  
		newGear = self.mrGbMCurrentGear + self.mrGbMRanges[self.mrGbMCurrentRange].downGearOffset
	end
	newGear = mrGearboxMogli.mrGbMGetNewEntry( self, self.mrGbMGears, newGear, "gear" )
	
	if      newRange ~= self.mrGbMCurrentRange
			and mrGearboxMogli.mrGbMSpeedLevelCheck( self, newGear, newRange, self.mrGbMCurrentRange2 ) then
		self:mbSetState11( "mrGbMCurrentRange", newRange, noEventSend ) 
		self:mbSetState11( "mrGbMCurrentGear",  newGear,  noEventSend ) 		
	end
end

function mrGearboxMogli:mrGbMSetRange2(new, noEventSend)
	if      not ( self.mrGbMAutoClutch )
			and self.mrGbMManualClutch > self.mrGbMClutchAfterShiftRanges2 + 0.1 
			and self.mrGbMManualClutch > self.mrGbMClutchPercent then
		return
	end		

	local newRange2 = mrGearboxMogli.mrGbMGetNewEntry( self, self.mrGbMRanges2, new, "range2" )
	
	if      newRange2 ~= self.mrGbMCurrentRange2
			and mrGearboxMogli.mrGbMSpeedLevelCheck( self, self.mrGbMCurrentGear, self.mrGbMCurrentRange, newRange2 ) then
		self:mbSetState11( "mrGbMCurrentRange2", newRange2, noEventSend ) 
	end
end

function mrGearboxMogli:mrGbMSetAutomatic( value, noEventSend )
	local new = false 
	if value and ( self.mrGbMAutoShiftUpRpm ~= nil or self.mrGbMAutoShiftDownRpm ~= nil ) then
		new = true 
	end 
	self:mbSetState11( "mrGbMAutomatic", new, noEventSend ) 		
end 

function mrGearboxMogli:mrGbMPrepareGearShift( timeToShift, clutchPercent, doubleClutch )
	if self.isServer then
		self.mrGbMGearShiftingNeeded  = 1 
		self.mrGbMGearShiftRpm        = self.mrGbMCurrentRpm
		self.mrGbMGearShiftingTime    = math.max( self.mrGbMGearShiftingTime, g_currentMission.time + timeToShift ) 
		self.mrGbMClutchPercent       = math.min( self.realClutchPercent, clutchPercent ) 
		self.mrGbMWheelRpmBeforeShift = self.mrGbMWheelRpm
		if not ( self.mrGbMAutoClutch ) then
			self.mrGbMDoubleClutch       = false
		elseif doubleClutch then
			self.mrGbMDoubleClutch       = true
			self.mrGbMClutchShiftingTime = math.max( self.mrGbMClutchShiftingTime, g_currentMission.time + 0.4 * timeToShift ) 
		else
			self.mrGbMDoubleClutch       = false
			self.mrGbMClutchShiftingTime = math.max( self.mrGbMClutchShiftingTime, g_currentMission.time + self.mrGbMClutchShiftTime ) 
		end
	end 
end

function mrGearboxMogli:mrGbMOnSetReverse( old, new, noEventSend )

	self.mrGbMReverseActive = new 
	--timer to shift the "reverse/forward"
	mrGearboxMogli.mrGbMPrepareGearShift( self, self.mrGbMGearTimeToShiftReverse, self.mrGbMClutchAfterShiftReverse, self.mrGbMReverseDoubleClutch ) 

	local default = self.mrGbMCurrentGear
	if self.mrGbMReverseResetGear then
		default = self.mrGbMDefaultGear
	end
	
	self:mbSetState11( "mrGbMDefaultGear", self.mrGbMCurrentGear, noEventSend ) 
	self:mbSetState11( "mrGbMCurrentGear", mrGearboxMogli.mrGbMGetNewEntry( self, self.mrGbMGears, default, "gear" ), noEventSend ) 
	
	local default = self.mrGbMCurrentRange
	if self.mrGbMReverseResetRange then
		default = self.mrGbMDefaultRange
	end

	self:mbSetState11( "mrGbMDefaultRange", self.mrGbMCurrentRange, noEventSend ) 
	self:mbSetState11( "mrGbMCurrentRange", mrGearboxMogli.mrGbMGetNewEntry( self, self.mrGbMRanges, default, "range" ), noEventSend ) 

	local default = self.mrGbMCurrentRange2
	if self.mrGbMReverseResetRange2 then
		default = self.mrGbMDefaultRange2
	end

	self:mbSetState11( "mrGbMDefaultRange2", self.mrGbMCurrentRange2, noEventSend ) 
	self:mbSetState11( "mrGbMCurrentRange2", mrGearboxMogli.mrGbMGetNewEntry( self, self.mrGbMRanges2, default, "range2" ), noEventSend ) 
	
	if self.isServer then
		self.mrGbMAutoShiftDownTimer = g_currentMission.time + self.mrGbMAutoShiftTimeout 
		self.mrGbMAutoShiftUpTimer   = g_currentMission.time + self.mrGbMAutoShiftTimeout 
	end 
end 
		
function mrGearboxMogli:mrGbMOnSetNeutral( old, new, noEventSend )		
	self.mrGbMNeutralActive = new 
	mrGearboxMogli.mrGbMPrepareGearShift( self, self.mrGbMGearTimeToShiftHl, self.mrGbMClutchAfterShiftHl ) 
end 

function mrGearboxMogli:mrGbMOnSetRange( old, new, noEventSend )
		
	self.mrGbMCurrentRange = new

	--timer to shift the "range"
	mrGearboxMogli.mrGbMPrepareGearShift( self, self.mrGbMGearTimeToShiftHl, self.mrGbMClutchAfterShiftHl, self.mrGbMRange1DoubleClutch ) 
	
	if self.isServer then
		self.mrGbMAutoShiftDownTimer = 0
		self.mrGbMAutoShiftUpTimer   = 0
	end 
end 

function mrGearboxMogli:mrGbMOnSetRange2( old, new, noEventSend )
		
	self.mrGbMCurrentRange2 = new

	--timer to shift the "range 2"
	mrGearboxMogli.mrGbMPrepareGearShift( self, self.mrGbMGearTimeToShiftRanges2, self.mrGbMClutchAfterShiftRanges2, self.mrGbMRange2DoubleClutch ) 
	
	if self.isServer then
		self.mrGbMAutoShiftDownTimer = 0
		self.mrGbMAutoShiftUpTimer   = 0
	end 
end 

function mrGearboxMogli:mrGbMOnSetGear( old, new, noEventSend )

	if     new > self.mrGbMCurrentGear then
		if self.isServer then
			self.mrGbMAutoShiftDownTimer = g_currentMission.time + self.mrGbMAutoShiftTimeout 
			self.mrGbMAutoShiftUpTimer   = g_currentMission.time + self.mrGbMAutoShiftTimeout2
		end 
	elseif new < self.mrGbMCurrentGear then                                              
		if self.isServer then
			self.mrGbMAutoShiftDownTimer = g_currentMission.time + self.mrGbMAutoShiftTimeout2
			self.mrGbMAutoShiftUpTimer   = g_currentMission.time + self.mrGbMAutoShiftTimeout 
		end 
	end			
	
-- adjust clutch % to fit rpm after gear shift	
	local clutchPercent = self.mrGbMClutchAfterShiftGear
	if clutchPercent < 0 then
		local gearSpeed = self.mrGbMGears[new].speed * self.mrGbMRanges[self.mrGbMCurrentRange].ratio * self.mrGbMRanges2[self.mrGbMCurrentRange2].ratio * self.mrGbMGlobalRatioFactor
		if self.mrGbMReverseActive then	
			gearSpeed = gearSpeed * self.mrGbMReverseRatio 
		end
		clutchPercent = Utils.clamp( mrGearboxMogli.mrGbMGetClutchPercent( self, self.mrGbMRatedRpm, self.mrGbMCurrentRpm, gearSpeed ), self.mrGbMMinClutchPercent, self.mrGbMMaxClutchPercent )
	end

	self.mrGbMCurrentGear = new
	
	--timer to set the gear
	mrGearboxMogli.mrGbMPrepareGearShift( self, self.mrGbMGearTimeToShiftGear, clutchPercent, self.mrGbMGearsDoubleClutch ) 	 	
end

function mrGearboxMogli:mrGbMOnSetIsOn( old, new, noEventSend )

	self.mrGbMIsOn = new 
	
	if self.isServer then
		if self.mrGbMIsOn then
				
			if      g_currentMission.driveControl ~= nil 
					and g_currentMission.driveControl.useModules ~= nil
					and g_currentMission.driveControl.useModules.shuttle
					and self.driveControl ~= nil
					and self.driveControl.shuttle ~= nil 
					and self.driveControl.shuttle.isActive then
				if self.driveControl.shuttle.direction < 0 then
					self:mbSetState11( "mrGbMReverseActive", true, noEventSend )  -- first gear, H range, reverse
				else
					self:mbSetState11( "mrGbMReverseActive", false, noEventSend )  -- first gear, H range, forward
				end 
			end
			self:mrGbMDoGearShift() 
		else
			self:mbSetState11( "mrGbMDefaultGear", self.mrGbMCurrentGear, noEventSend ) 
			self:mbSetState11( "mrGbMDefaultRange", self.mrGbMCurrentRange, noEventSend ) 
			self:mbSetState11( "mrGbMDefaultRange2", self.mrGbMCurrentRange2, noEventSend ) 
			self.mrGbMGearShiftingNeeded   = false 
		  self.motor.minRpm              = self.mrGbMMotorMinRpm
		  self.motor.maxRpm              = self.mrGbMMotorMaxRpm
			self.motor.lowBrakeForceScale	 = self.mrGbMLowBrakeForceScale
			self.motor.torqueCurve         = self.mrGbMTorqueCurveBackup
			self.motor.maxClutchTorque     = self.mrGbMMaxClutchTorque
			self.motor.ptoMotorRpmRatio    = self.mrGbMPtoMotorRpmRatio
		end 
	end 
	
end 

function mrGearboxMogli:mrGbMOnSetWarningText( old, new, noEventSend )
	self.mrGbMWarningText  = new
  self.mrGbMWarningTimer = g_currentMission.time + 2000
end

function mrGearboxMogli:newGetRpmLimit( superFunc, ... )
	if self.mrGbMIsOn then
		local rpmLimit = superFunc( self, ... )
		return math.min( rpmLimit, self.mrGbMMaxPossibleRpm )		
	else
		return superFunc( self, ... )
	end
end


function mrGearboxMogli:newUpdateVehiclePhysics(superFunc, axisForward, axisForwardIsAnalog, axisSide, axisSideIsAnalog, dt)

	if self.mrGbMIsOn then
		if      g_currentMission.driveControl ~= nil 
				and g_currentMission.driveControl.useModules ~= nil
				and g_currentMission.driveControl.useModules.shuttle
				and self.driveControl ~= nil
				and self.driveControl.shuttle ~= nil 
				and self.driveControl.shuttle.isActive then
			self.mrGbMLastAxisForward = self.driveControl.shuttle.direction * axisForward
		else
			self.mrGbMLastAxisForward = axisForward

			local wantedDirection = 1
			if self.mrGbMReverseActive then
				wantedDirection = -1
			end;
			
			if      axisForward        >= 0 
					and self.mrGbMLastSpeed < 0.5 then
			-- hand brake 
				self.motor.lowBrakeForceScale = mrGearboxMogli.huge
			elseif  axisForward        > 0.01
					and self.movingDirection ~= wantedDirection then
			-- wrong direction while braking
				self.motor.lowBrakeForceScale = mrGearboxMogli.huge
			elseif  axisForward        > 0.01 then
			-- braking 
				self.motor.lowBrakeForceScale = 2 * self.mrGbMLowBrakeForceScale
			elseif self.mrGbMNoTransmission   then
			-- neutral or gear shift 
				self.motor.lowBrakeForceScale = 0
			else 
			-- accelerating 
				self.motor.lowBrakeForceScale = self.mrGbMRealMotorBrakeFx
			end
					
			if self.movingDirection ~= wantedDirection or self.mrGbMLastSpeed < 0.1 then
				axisForward = math.min( 0, axisForward ) 	
			end
			
			if self.mrGbMReverseActive then
				axisForward = -axisForward;
			end;
			
			self.GPSmovingDirection    = wantedDirection
			self.GPSmovingDirectionCnt = 40; --here we remotely handle direction mechanics from GPS. no problem if GPS is not installed
		end

		if self.mrGbMLastAxisForward <= 0 then
			if self.mrGbMLastAxisForward > - self.mrGbMLastMinThrottle then
				if self.mrGbMReverseActive then
					axisForward       = self.mrGbMLastMinThrottle
				else
					axisForward       = -self.mrGbMLastMinThrottle
				end
			--axisForwardIsAnalog = true
			end
		end
	end
	
	superFunc(self, axisForward, axisForwardIsAnalog, axisSide, axisSideIsAnalog, dt)
end;

function mrGearboxMogli:newGetTorque( superFunc, ... )
	if self.vehicle.mrGbMIsOn then
		local t, b = superFunc( self, ... )
		--if      self.nonClampedMotorRpm < self.minRpm then
		--	print(tostring(t))
		--	t = t * math.max( self.nonClampedMotorRpm / self.minRpm, 0.1 )
		--end
		if      self.vehicle.mrGbMLastSlipFactor ~= nil 
				and not ( self.vehicle.mrGbMNoTransmission ) 
				and self.vehicle.realClutchPercent >= 1E-3 then
			t = t * ( self.vehicle.mrGbMTransmissionEfficiency - self.vehicle.mrGbMTransmissionEfficiencyDec * ( 1 - self.vehicle.mrGbMLastSlipFactor ) )
		end
		self.mrGbMLastMotorTorque = t
		return t, b
	end
	return superFunc( self, ... )
end

function mrGearboxMogli:newGetBestGearRatio( superFunc, ... )
	if self.vehicle.mrGbMIsOn then
		local maxWheelRpm = math.max( self.vehicle.mrGbMCurrentGearSpeed * 9.5492965855137201461330258023509, 1E-3 )
		if maxWheelRpm > 1E-5 then 
			return math.min( self.vehicle.mrGbMRatedRpm / maxWheelRpm, mrGearboxMogli.huge )
		else
			return mrGearboxMogli.huge 
		end
	end
	return superFunc( self, ... )
end


Drivable.getRpmLimit          = Utils.overwrittenFunction( Drivable.getRpmLimit,          mrGearboxMogli.newGetRpmLimit )
Drivable.updateVehiclePhysics = Utils.overwrittenFunction( Drivable.updateVehiclePhysics, mrGearboxMogli.newUpdateVehiclePhysics )
VehicleMotor.getTorque        = Utils.overwrittenFunction( VehicleMotor.getTorque,        mrGearboxMogli.newGetTorque )
VehicleMotor.getBestGearRatio = Utils.overwrittenFunction( VehicleMotor.getBestGearRatio, mrGearboxMogli.newGetBestGearRatio )
